;;;
;;; Universal Monitor 8086
;;;   Copyright (C) 2020 Haruo Asano
;;;

	CPU	8086

TARGET:	EQU	"8086"

	INCLUDE	"config.inc"

	INCLUDE "../common.inc"


	ASSUME	CS:CODE, DS:DATA, SS:DATA, ES:NOTHING

;;;
;;; ROM area
;;;

	SEGMENT	CODE
	
	ORG	ROM_B

CSTART:
	MOV	AX,RAMSEG
	MOV	DS,AX
	MOV	SS,AX
	MOV	SP,STACK

	;; 
	
	CALL	INIT

	MOV	AX,RAMSEG
	MOV	[CURSEG],AX
	IF RAMSEG = 0
	MOV	AX,VECSIZ*4
	ELSE
	XOR	AX,AX
	ENDIF
	MOV	[DSADDR],AX
	MOV	[GADDR],AX
	MOV	[SADDR],AX
	MOV	AL,'I'
	MOV	[HEXMOD],AL
	XOR	AL,AL
	MOV	[PSPEC],AL

	;; Initialize vector
	XOR	AX,AX
	MOV	ES,AX
	MOV	SI,INIVEC
	MOV	DI,AX
	MOV	CX,VECSIZ*2
INI0:	
	MOV	AX,CS:[SI]
	INC	SI
	INC	SI
	MOV	ES:[DI],AX
	INC	DI
	INC	DI
	DEC	CX
	JZ	INI2
	CMP	SI,INIVECE
	JNZ	INI0
	MOV	AX,INTDMY
	MOV	BX,ROMSEG
INI1:
	MOV	ES:[DI],AX
	INC	DI
	INC	DI
	MOV	ES:[DI],BX
	INC	DI
	INC	DI
	DEC	CX
	JNZ	INI1
INI2:
	
	;; Opening message
	MOV	BX,OPNMSG
	CALL	MSGOUT

WSTART:
	MOV	BX,PROMPT
	CALL	MSGOUT
	CALL	GETLIN
	MOV	BX,INBUF
	CALL	SKIPSP
	CALL	UPPER
	AND	AL,AL
	JZ	WSTART

	CMP	AL,'D'
	JNE	M00
	JMP	DUMP
M00:	
	CMP	AL,'G'
	JNE	M01
	JMP	GO
M01:
	CMP	AL,'S'
	JNE	M02
	JMP	SETM
M02:	

ERR:
	MOV	BX,ERRMSG
	CALL	MSGOUT
	JMP	WSTART

;;;
;;; DUMP memory
;;;

DUMP:
	INC	BX
	CALL	SKIPSP
	CALL	RDHEX
	OR	CL,CL
	JNZ	DP0
	;; No arg.
	CALL	SKIPSP
	OR	AL,AL
	JNZ	ERR

	MOV	DX,[DSADDR]
	ADD	DX,128
	MOV	[DEADDR],DX
	JMP	DPM

	;; 1st arg. found
DP0:
	MOV	AL,[BX]
	CMP	AL,':'
	JNZ	DP01
	MOV	[CURSEG],DX
	INC	BX
	CALL	RDHEX
	OR	CL,CL
	JZ	ERR
DP01:	
	MOV	[DSADDR],DX
	MOV	AL,[BX]
	CMP	AL,','
	JZ	DP1
	OR	AL,AL
	JNZ	ERR
	;; No 2nd arg.
	ADD	DX,128
	MOV	[DEADDR],DX
	JMP	DPM
DP1:
	INC	BX
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	OR	CL,CL
	JZ	ERR
	MOV	AL,[BX]
	OR	AL,AL
	JNZ	ERR
	INC	DX
	MOV	[DEADDR],DX
DPM:
	;; DUMP main
	MOV	ES,[CURSEG]
	MOV	SI,[DSADDR]
	AND	SI,0FFF0H
	XOR	AL,AL
	MOV	[DSTATE],AL
DPM0:
	CALL	DPL
	CALL	CONST
	OR	AL,AL
	JNZ	DPM1
	MOV	AL,[DSTATE]
	CMP	AL,2
	JC	DPM0
	MOV	BX,[DEADDR]
	MOV	[DSADDR],BX
	JMP	WSTART
DPM1:
	MOV	[DSADDR],SI
	CALL	CONIN
	JMP	WSTART

DPL:
	;; DUMP	line
	MOV	AX,[CURSEG]
	CALL	HEXOUT4
	MOV	AL,':'
	CALL	CONOUT
	MOV	AX,SI
	CALL	HEXOUT4
	MOV	BX,DSEP0
	CALL	MSGOUT
	MOV	DI,INBUF
	MOV	CX,16
DPL0:
	CALL	DPB
	LOOP	DPL0

	MOV	BX,DSEP1
	CALL	MSGOUT

	MOV	DI,INBUF
	MOV	CX,16
DPL1:
	MOV	AL,[DI]
	INC	DI
	CMP	AL,' '
	JC	DPL2
	CMP	AL,7FH
	JNC	DPL2
	CALL	CONOUT
	JMP	DPL3
DPL2:
	MOV	AL,'.'
	CALL	CONOUT
DPL3:
	LOOP	DPL1
	JMP	CRLF

DPB:
	;; DUMP byte
	MOV	AL,' '
	CALL	CONOUT
	MOV	AL,[DSTATE]
	OR	AL,AL
	JNZ	DPB2
	;; State 0
	CMP	SI,[DSADDR]
	JZ	DPB1
DPB0:
	;; Still 0 or 2
	MOV	AL,' '
	MOV	[DI],AL
	CALL	CONOUT
	MOV	AL,' '
	CALL	CONOUT
	INC	SI
	INC	DI
	RET
DPB1:
	;; Found start address
	MOV	AL,1
	MOV	[DSTATE],AL
DPB2:
	MOV	AL,[DSTATE]
	CMP	AL,1
	JNZ	DPB0		; state 2
	;; DUMP state 1
	MOV	AL,ES:[SI]
	MOV	[DI],AL
	CALL	HEXOUT2
	INC	SI
	INC	DI

	CMP	SI,[DEADDR]
	JNZ	DPB3
	;; Found end address
	MOV	AL,2
	MOV	[DSTATE],AL
DPB3:
	RET

;;;
;;; GO address
;;; 

GO:
	INC	BX
	CALL	SKIPSP
	CALL	RDHEX
	OR	CL,CL
	JZ	G1
	MOV	AL,[BX]
	CMP	AL,':'
	JNE	G0
	MOV	[CURSEG],DX
	INC	BX
	CALL	RDHEX
	OR	CL,CL
	JZ	GE
G0:
	MOV	[GADDR],DX
	MOV	AL,[BX]
	OR	AL,AL
	JNZ	GE
G1:
	PUSH	[CURSEG]
	PUSH	[GADDR]
	RETF
GE:
	JMP	ERR

;;;
;;; SET memory
;;;

SETM:
	INC	BX
	CALL	SKIPSP
	CALL	RDHEX
	OR	CL,CL
	JZ	SM1
	MOV	AL,[BX]
	CMP	AL,':'
	JNE	SM0
	MOV	[CURSEG],DX
	INC	BX
	CALL	RDHEX
	OR	CL,CL
	JZ	GE
SM0:
	MOV	[SADDR],DX
	CALL	SKIPSP
	MOV	AL,[BX]
	OR	AL,AL
	JNZ	GE
SM1:
	MOV	AX,[CURSEG]
	MOV	ES,AX
	CALL	HEXOUT4
	MOV	AL,':'
	CALL	CONOUT
	MOV	SI,[SADDR]
	MOV	AX,SI
	CALL	HEXOUT4
	MOV	BX,DSEP1
	CALL	MSGOUT
	MOV	AL,ES:[SI]
	CALL	HEXOUT2
	MOV	AL,' '
	CALL	CONOUT
	CALL	GETLIN
	MOV	BX,INBUF
	CALL	SKIPSP
	MOV	AL,[BX]
	OR	AL,AL
	JNZ	SM2
	;; Empty (Increment address)
	INC	SI
	MOV	[SADDR],SI
	JMP	SM1
SM2:
	CMP	AL,'-'
	JNZ	SM3
	;; '-' (Decrement address)
	DEC	SI
	MOV	[SADDR],SI
	JMP	SM1
SM3:
	CMP	AL,'.'
	JNZ	SM4
	;; '.' (Quit)
	JMP	WSTART
SM4:
	CALL	RDHEX
	OR	CL,CL
	JZ	SME
	MOV	ES:[SI],DL
	INC	SI
	MOV	[SADDR],SI
	JMP	SM1
SME:
	JMP	ERR

;;;
;;; Other support routines
;;;

MSGOUT:
	MOV	AL,CS:[BX]
	AND	AL,AL
	JZ	STROE
	CALL	CONOUT
	INC	BX
	JMP	STROUT
MSGOE:
	RET

STROUT:
	MOV	AL,[BX]
	AND	AL,AL
	JZ	STROE
	CALL	CONOUT
	INC	BX
	JMP	STROUT
STROE:
	RET

HEXOUT4:
	PUSH	AX
	MOV	AL,AH
	CALL	HEXOUT2
	POP	AX
HEXOUT2:
	PUSH	AX
	SHR	AL,1
	SHR	AL,1
	SHR	AL,1
	SHR	AL,1
	CALL	HEXOUT1
	POP	AX
HEXOUT1:
	AND	AL,0FH
	ADD	AL,'0'
	CMP	AL,'9'+1
	JC	HO10
	ADD	AL,'A'-'9'-1
HO10:
	JMP	CONOUT

CRLF:
	MOV	AL,CR
	CALL	CONOUT
	MOV	AL,LF
	JMP	CONOUT

GETLIN:
	MOV	BX,INBUF
	XOR	CL,CL
GL0:
	CALL	CONIN
	CMP	AL,CR
	JZ	GLE
	CMP	AL,LF
	JZ	GLE
	CMP	AL,BS
	JZ	GLB
	CMP	AL,DEL
	JZ	GLB
	CMP	AL,' '
	JC	GL0
	CMP	AL,80H
	JNC	GL0
	CMP	CL,BUFLEN-1
	JNC	GL0		; Too long
	MOV	[BX],AL
	CALL	CONOUT
	INC	BX
	INC	CL
	JMP	GL0
GLB:
	AND	CL,CL
	JZ	GL0
	DEC	BX
	DEC	CL
	MOV	AL,BS
	CALL	CONOUT
	MOV	AL,' '
	CALL	CONOUT
	MOV	AL,BS
	CALL	CONOUT
	JMP	GL0
GLE:
	CALL	CRLF
	MOV	AL,00H
	MOV	[BX],AL
	RET
	
SKIPSP:
	MOV	AL,[BX]
	CMP	AL,' '
	JNE	SSR
	INC	BX
	JMP	SKIPSP
SSR:
	RET

UPPER:
	CMP	AL,'a'
	JC	UPR
	CMP	AL,'z'+1
	JNC	UPR
	ADD	AL,'A'-'a'
UPR:
	RET

RDHEX:
	XOR	CL,CL
	XOR	DX,DX
RH0:
	MOV	AL,[BX]
	CALL	UPPER
	CMP	AL,'0'
	JC	RHE
	CMP	AL,'9'+1
	JC	RH1
	CMP	AL,'A'
	JC	RHE
	CMP	AL,'F'+1
	JNC	RHE
	SUB	AL,'A'-'9'-1
RH1:
	SUB	AL,'0'
	SHL	DX,1
	SHL	DX,1
	SHL	DX,1
	SHL	DX,1
	OR	DL,AL
	INC	BX
	INC	CL
	JMP	RH0
RHE:
	RET

;;;
;;; Interrupt Handler
;;;

INTDMY:
	IRET
	
OPNMSG:
	DB	CR,LF,"Universal Monitor 8086",CR,LF,00H

PROMPT:
	DB	"] ",00H

IHEMSG:
	DB	"Error ihex",CR,LF,00H
SHEMSG:
	DB	"Error srec",CR,LF,00H
ERRMSG:
	DB	"Error",CR,LF,00H

DSEP0:
	DB	" :",00H
DSEP1:
	DB	" : ",00H
IHEXER:
        DB	":00000001FF",CR,LF,00H
SRECER:
        DB	"S9030000FC",CR,LF,00H

INIVEC:
	DW	INTDMY, ROMSEG	; 00H Divide error
	DW	INTDMY, ROMSEG	; 01H Single step
	DW	INTDMY, ROMSEG	; 02H NMI
	DW	INTDMY, ROMSEG	; 03H Breakpoint
	DW	INTDMY, ROMSEG	; 04H Overflow
INIVECE:
	
	IF USE_DEV_8251
	INCLUDE	"dev/dev_8251.asm"
	ENDIF

	IF USE_DEV_EMILY
	INCLUDE	"dev/dev_emily.asm"
	ENDIF

	;;
	;; Reset entry point
	;;
	
	ORG	( 0FFFF0H - ROMSEG * 16 )

	JMPF	ROMSEG:CSTART
	
;;;
;;; RAM area
;;;

	SEGMENT	DATA

	;;
	;; Work area
	;;

	ORG	WORK_B

INBUF:	DS	BUFLEN	; Line input buffer
CURSEG:	DS	2	; Current segment
DSADDR:	DS	2	; DUMP start address
DEADDR:	DS	2	; DUMP end address
DSTATE:	DS	1	; DUMP state
	ALIGN	2
GADDR:	DS	2	; GO address
SADDR:	DS	2	; SET address
HEXMOD:	DS	1	; HEX file mode
PSPEC:	DS	1	; Processor spec.

RECTYP:	DS	1	; Record type
	
	END
	
